import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils import check_array, check_X_y
from sklearn.utils.estimator_checks import check_estimator
from sklearn.utils.validation import check_is_fitted


# TODO this needs to be savable

def get_default_estimators():
    print('---- get default ests')

class BaseStacker(object):
    def _model_predict(self, model, X):
        try:
            return model.predict_proba(X)[:,1]
        except AttributeError:
            return model.predict(X)

    def _predict_proba(self, X):
        return np.array([
            self._model_predict(model, X)
            for model in self._estimators
        ]).T

    def fit(self, X, y=None):
        X, y = check_X_y(X, y)
        #y = self._lb.fit_transform(y)

        for model in self._estimators:
            model.fit(X, y)

        results = self._predict_proba(X)
        self._stack_estimator.fit(results, y)
        self.fitted_ = True
        return self

class StackingRegressor(BaseEstimator, RegressorMixin, BaseStacker):

    def __init__(self, estimators=None, stack_estimator=None):
        self.estimators = estimators
        self.stack_estimator = stack_estimator
        self._lb = LabelBinarizer()

        if self.estimators is None:
            import xgboost as xgb
            from sklearn.ensemble import RandomForestRegressor
            from sklearn.linear_model import LogisticRegression, BayesianRidge, SGDRegressor
            self._estimators = [
                xgb.sklearn.XGBRegressor(max_depth=50, n_estimators=300),
                BayesianRidge(),
                RandomForestRegressor(n_estimators=300),
                SGDRegressor(),
            ]
        else:
            self._estimators = self.estimators

        if self.stack_estimator is None:
            self._stack_estimator = xgb.sklearn.XGBRegressor(max_depth=50, n_estimators=1000)
        else:
            self._stack_estimator = self.stack_estimator

    def predict(self, X):
        check_is_fitted(self, 'fitted_')
        X = check_array(X)
        results = self._predict_proba(X)
        return self._stack_estimator.predict(results)

class StackingClassifier(BaseEstimator, ClassifierMixin, BaseStacker):

    def __init__(self, estimators=None, stack_estimator=None):
        self.estimators = estimators
        self.stack_estimator = stack_estimator
        self._lb = LabelBinarizer()

        if self.estimators is None:
            import xgboost as xgb
            from sklearn.ensemble import RandomForestClassifier
            from sklearn.linear_model import LogisticRegression, BayesianRidge, SGDClassifier
            self._estimators = [
                xgb.sklearn.XGBClassifier(max_depth=50, n_estimators=300),
                BayesianRidge(),
                LogisticRegression(max_iter=2000),
                RandomForestClassifier(n_estimators=300),
                SGDClassifier(),
            ]
        else:
            self._estimators = self.estimators

        if self.stack_estimator is None:
            self._stack_estimator = xgb.sklearn.XGBClassifier(max_depth=50, n_estimators=1000)
        else:
            self._stack_estimator = self.stack_estimator


    def predict(self, X):
        check_is_fitted(self, 'fitted_')
        X = check_array(X)
        results = self._predict_proba(X)
        return self._stack_estimator.predict(results)

    def predict_proba(self, X):
        check_is_fitted(self, 'fitted_')
        X = check_array(X)
        results = self._predict_proba(X)
        return self._stack_estimator.predict_proba(results)


# TODO want an online kaggle ensembler....
#
# should take many features/transforms
# runs as long as you tell it too, continually running with cv & test set....
#
# should sample features, sample model, train, then add results to ensemble
#
# should be saveable
#
# should be online to allow runtime updates and improvements
#



import numpy as np

import pandas as pd
from scipy import stats


def boxcox(y, ld=None, offset=None):
    if offset is None:
        offset = abs(y.min()) + 1
    if ld is None:
        yt, ld = stats.boxcox(y + offset)
    else:
        yt = stats.boxcox(y + offset, ld)
    return pd.Series(yt), ld, offset


def invboxcox(y, ld, offset=0):
    if ld == 0:
        return np.exp(y) - offset
    else:
        return np.exp(np.log(ld*y+1)/ld) - offset



from unittest import TestCase

from corgi.dictionary import rename_keys


class TestDictionaryFunctions(TestCase):

    def test_rename_keys(self):
        self.assertEqual(
            rename_keys(
                {'one': 1, 'two': 2},
                {'one': '1', 'two': '2'}
            ),
            {'1': 1, '2': 2, }
        )



from unittest import TestCase

from sklearn.utils.estimator_checks import check_estimator

from ..ensembling import StackingClassifier, StackingRegressor


class TestEnsembling(TestCase):
    def test_check_regressor(self):
        check_estimator(StackingRegressor)

    def test_check_classifier(self):
        check_estimator(StackingClassifier)



import numpy as np


def remove_low_variance(X, low_variance_threshold=1e-4, axis=0):
    low_variance = X.var(axis=axis) < low_variance_threshold
    X[:, ~low_variance]
    return X


def normalize(X):
    zero_mean = X - X.mean()
    return (zero_mean + zero_mean.min()) / X.max()


def independent_columns(A, tol=1e-05):
    """
    Return an array composed of independent columns of A.

    Note the answer may not be unique; this function returns one of many
    possible answers.

    http://stackoverflow.com/q/13312498/190597 (user1812712)
    http://math.stackexchange.com/a/199132/1140 (Gerry Myerson)
    http://mail.scipy.org/pipermail/numpy-discussion/2008-November/038705.html
        (Anne Archibald)

    >>> A = np.array([(2,4,1,3),(-1,-2,1,0),(0,0,2,2),(3,6,2,5)])
     2  4  1  3
    -1 -2  1  0
     0  0  2  2
     3  6  2  5
    # try with checking the rank of matrices
    >>> independent_columns(A)
    np.array([[ 2, 1],
              [-1, 1],
              [ 0, 2],
              [ 3, 2]]), np.array([0, 2])
    """
    # TODO need a source for diag(R) > 0, being the independent indices of A
    Q, R = np.linalg.qr(A)
    independent = np.where(np.abs(R.diagonal()) > tol)[0]
    return A[:, independent], independent



import numpy as np

import pandas as pd
from scipy.stats import kendalltau, spearmanr
from sklearn.metrics import (accuracy_score, f1_score, log_loss,
                             mean_squared_error, precision_score, recall_score)
from sklearn.model_selection import StratifiedKFold
from tqdm import tqdm


classifier_scoring = {
    'accuracy': accuracy_score,
    'log_loss': log_loss,
    'f1_score': lambda x, y: f1_score(x, y, average='weighted'),
    'precision': lambda x, y: precision_score(x, y, average='weighted'),
    'recall': lambda x, y: recall_score(x, y, average='weighted'),
}

regression_scoring = {
    'mean_squared_error': mean_squared_error,
    'kendalltau': lambda x, y: kendalltau(x, y).correlation,
    'spearmanr': lambda x, y: spearmanr(x, y)[0],
}

def scores(y, y_pred, scoring=None):
    if scoring is None:
        raise Exception("cross_val_scores requires a dict of measures.")

    scores = {}
    for k, metric in scoring.items():
        scores[k] = metric(y, y_pred)
    return scores


def cross_val_scores(clf, X, y, cv=3, scoring=None):
    if scoring is None:
        raise Exception("cross_val_scores requires a dict of measures.")

    X, y = np.array(X), np.array(y)
    skf = StratifiedKFold(n_splits=cv)
    scores = []
    for train, test in tqdm(skf.split(X, y)):
        clf.fit(X[train], y[train])
        y_pred = clf.predict(X[test])
        score = {}
        for k, metric in scoring.items():
            try:
                score[k] = metric(y[test], y_pred)
            except:
                pass
        scores.append(score)

    return pd.DataFrame(scores)



import math

import numpy as np

import pandas as pd

one_hot = lambda df: pd.get_dummies(df)
zero_mean = lambda df: df - df.mean(axis=0)
one_dev = lambda df: df / df.std(axis=0)
no_nan = lambda df: df.fillna(0)
shuffled = lambda df: df.sample(frac=1)
remove_constant_cols = lambda df: df.loc[:, (df != df.iloc[0]).any()]

def date_features(df):
    df = df.apply(pd.to_datetime)
    date_features = [
        'month', 'day', 'dayofyear',
        'week', 'weekday', 'quarter',
        'year', 'is_month_start', 'is_month_end',
        'is_quarter_start', 'is_quarter_end', 'is_year_start',
        'is_year_end'
    ]

    cols = df.columns

    for date_feature in date_features:
        new_cols = [c + '_' + date_feature for c in cols]
        df[new_cols] = df[cols].apply(
            lambda x: getattr(x.dt, date_feature)
        )

    # we've extracted some potentially useful date features, lets convert
    # the original dates to seconds so they can be used as a numeric feature
    df[cols] = df[cols].astype(np.int64)

    return df

def clean(df, drop_cols=None, categorical_cols=None, date_cols=None):
    cols = df.columns

    if date_cols is not None:
        df = pd.concat(
            [
                df.drop(date_cols, axis=1),
                date_features(df[date_cols]),
            ],
            axis=1
        )

    if categorical_cols is not None:
        df = pd.concat(
            [
                df.drop(categorical_cols, axis=1),
                one_hot(df[categorical_cols].astype(object))
            ],
            axis=1
        )

    # Drop unneeded and non-informative
    if drop_cols is not None:
        df = df.drop(drop_cols, axis=1)

    df = remove_constant_cols(df)

    return df



def rename_keys(dictionary, names_map):
    for k, v in names_map.iteritems():
        if k not in dictionary:
            continue
        dictionary[v] = dictionary[k]
        del dictionary[k]
    return dictionary



def listify(obj):
    if isinstance(obj, list) or isinstance(obj, tuple):
        return obj
    return [obj]


def dictify(obj, key):
    if isinstance(obj, dict):
        return obj
    return {key: obj}



from __future__ import division, print_function

import numpy as np


def hist(data, width=20):
    counts, values = np.histogram(data)
    max_count = counts.max()

    for (count, value) in zip(counts, values):
        scaled = int(round((count / max_count) * width))
        print('%5.2f (%5d):' % (value, count), 'X'*scaled)

def sql_hist(table, column, where=''):
    """
    Queries a table, selecting the column and count of the values in that column. Generates
    a histogram plot of that data for display in a jupyter notebook.
    """

    # this depends on the type of table we're looking at, mysql, postgres, etc...
    # might want to template these queries
    sql = """
    SELECT something FROM %(table)s
    %(where)s
    """
    pass

def sql_scatter(table, x_col, y_col, coloring_dimension=None, where=''):
    """
    Queries a table for data in x_col and y_col, then displays them (optionally colored by
    coloring_dimension) onto a scatterplot.
    """
    pass
